# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

"""
This module contains validation schema for config file loaded values
"""

from schema import Schema, And, Optional, Use, SchemaError

from ..internal.exceptions import SSAError

configuration_schema = Schema({
    Optional('domains_number'): And(int, lambda n: n >= 0,
                                    error='number of domains should be a positive integer'),
    Optional('urls_number'): And(int, lambda n: n > 0,
                                 error='number of URLs should be a positive integer'),
    Optional('requests_duration'): And(Use(lambda n: float(n)),
                                       lambda n: n > 0.0,
                                       error='request duration should be positive'),
    Optional('request_number'): And(int, lambda n: n > 0,
                                    error='number of requests should be a positive integer'),
    Optional('time'): And(int, lambda n: 0 <= n <= 12,
                          error='time should be a positive integer between 1 and 12, or 0'),
    Optional('correlation'): And(bool,
                                 error='bad correlation value, should be On or Off'),
    Optional('correlation_coefficient'): And(Use(lambda n: float(n)),
                                             lambda n: 0.0 < n <= 1.0,
                                             error='correlation coefficient could not be negative or greater than 1'),
    Optional('ignore_list'): Use(lambda s: str(s)),
    Optional('summary_notification_enabled'): And(bool,
                                                  error='bad notification value, should be On or Off'),
})


def validate(data: dict, scheme: Schema = configuration_schema) -> dict:
    """
    Validate given data.
    Returns validated data
    """
    try:
        return scheme.validate(data)
    except SchemaError as e:
        raise SSAError(f'Invalid configuration: {str(e)}')
